import json
import logging
import math
import time
from collections import defaultdict
from typing import List, Optional, Callable, Union, Dict

from autobahn.exception import Disconnected
from autobahn.twisted.websocket import connectWS, WebSocketClientProtocol, WebSocketClientFactory
from twisted.internet.protocol import ReconnectingClientFactory

from ezbeq.apis.ws import WsServer
from ezbeq.catalogue import CatalogueProvider, CatalogueEntry
from ezbeq.device import SlotState, DeviceState, PersistentDevice

BEQ_FILTER_NAME_PATTERN = r'^BEQ_(Gain_\d+|\d+_[a-zA-Z0-9]+)$'

SLOT_ID = 'CamillaDSP'

logger = logging.getLogger('ezbeq.camilladsp')


class CamillaDspSlotState(SlotState):

    def __init__(self):
        super().__init__(SLOT_ID)
        self.gains: List[dict] = []
        self.mutes: List[dict] = []

    def as_dict(self) -> dict:
        return {
            **super().as_dict(),
            'gains': self.gains,
            'mutes': self.mutes
        }


class CamillaDspState(DeviceState):

    def __init__(self, name: str):
        self.__name = name
        self.__has_volume = False
        self.slot = CamillaDspSlotState()
        self.slot.active = True
        self.master_volume: float = 0.0
        self.mute: bool = False

    @property
    def has_volume(self) -> bool:
        return self.__has_volume

    @has_volume.setter
    def has_volume(self, has_volume: bool):
        self.__has_volume = has_volume

    def serialise(self) -> dict:
        val = {
            'type': 'camilladsp',
            'name': self.__name,
            'mute': self.mute,
            'has_volume': self.has_volume,
            'slots': [self.slot.as_dict()]
        }
        if self.has_volume is True:
            val['masterVolume'] = self.master_volume
        return val


class CamillaDsp(PersistentDevice[CamillaDspState]):

    def __init__(self, name: str, config_path: str, cfg: dict, ws_server: WsServer, catalogue: CatalogueProvider):
        super().__init__(config_path, name, ws_server)
        self.__name = name
        self.__catalogue = catalogue
        self.__ip: str = cfg['ip']
        self.__port: int = cfg['port']
        self.__beq_channels: List[int] = [int(c) for c in cfg['channels']]
        self.__current_config: dict = {}
        self.__levels_interval_ms = round(1000.0 / float(cfg.get('levelsFps', 10)))
        self.__config_updater: Optional[UpdateConfig] = None
        if not self.__beq_channels:
            raise ValueError(f'No channels supplied for CamillaDSP {name} - {self.__ip}:{self.__port}')
        self.__ws_client = cfg['make_wsclient'](self.__ip, self.__port, self)
        self.__playback_peak: List[float] = []
        self.__playback_rms: List[float] = []
        ws_server.factory.set_levels_provider(name, self.start_broadcast_levels)

    @property
    def current_config(self) -> dict:
        return self.__current_config

    @current_config.setter
    def current_config(self, cfg: dict):
        self.__current_config = cfg

        def upd():
            prev = self._current_state.has_volume
            has_filters = 'filters' in cfg and cfg['filters']
            if has_filters:
                vol_filter_name = next((k for k, v in cfg['filters'].items() if v['type'] == 'Volume'), None)
                if vol_filter_name and 'pipeline' in cfg:
                    self._current_state.has_volume = any(f for f in cfg['pipeline']
                                                         if f['type'] == 'Filter' and vol_filter_name in f['names'])
            if prev != self._current_state.has_volume:
                logger.info(f'[{self.name}] current config has volume filter? {self._current_state.has_volume}')

            if 'pipeline' in cfg and cfg['pipeline']:
                gains = [{'id': c, 'value': 0.0} for c in self.__beq_channels]
                mutes = [{'id': c, 'value': False} for c in self.__beq_channels]
                filters = cfg['filters'] if has_filters else {}
                for k, v in filters.items():
                    if k.startswith('BEQ_Gain_'):
                        channel = int(k[9:])
                        muted = v.get('mute', False)
                        gain = v.get('gain', 0.0)
                        if not math.isclose(gain, 0.0) or muted is True:
                            for f in cfg['pipeline']:
                                if f['type'] == 'Filter' and f['channel'] in self.__beq_channels and 'BEQ_Gain' in f['names']:
                                    next(g for g in gains if g['id'] == f['channel'])['value'] = gain
                                    next(g for g in mutes if g['id'] == f['channel'])['value'] = muted
                self._current_state.slot.gains = gains
                self._current_state.slot.mutes = mutes

        self._hydrate_cache_broadcast(upd)

    def start_broadcast_levels(self) -> None:
        from twisted.internet import reactor
        sched = lambda: reactor.callLater(self.__levels_interval_ms / 1000.0, __send)

        def __send():
            self.request_levels()
            if self.ws_server.has_levels_client(self.name):
                sched()

        sched()

    def _load_initial_state(self) -> CamillaDspState:
        return CamillaDspState(self.name)

    def _merge_state(self, loaded: CamillaDspState, cached: dict) -> CamillaDspState:
        if 'has_volume' in cached:
            loaded.has_volume = cached['has_volume']
        if 'masterVolume' in cached:
            loaded.master_volume = cached['masterVolume']
        if 'slots' in cached:
            for slot in cached['slots']:
                if 'id' in slot:
                    if slot['id'] == SLOT_ID:
                        if slot['last']:
                            loaded.slot.last = slot['last']
                        if slot['author']:
                            loaded.slot.last_author = slot['author']
                        if slot['gains']:
                            loaded.slot.gains = slot['gains']
                        if slot['mutes']:
                            loaded.slot.mutes = slot['mutes']
        return loaded

    @property
    def device_type(self) -> str:
        return self.__class__.__name__.lower()

    def update(self, params: dict) -> bool:
        def __do_it() -> bool:
            any_update = False
            if 'slots' in params:
                for slot in params['slots']:
                    if slot['id'] == SLOT_ID:
                        if 'entry' in slot:
                            if slot['entry']:
                                match = self.__catalogue.find(slot['entry'])
                                if match:
                                    mv = 0.0
                                    if 'gains' in slot and slot['gains']:
                                        mv = float(slot['gains'][0]['value'])
                                    self.load_filter(SLOT_ID, match, mv)
                                    any_update = True
                            else:
                                self.clear_filter(SLOT_ID)
                                any_update = True
                        elif 'gains' in slot or 'mutes' in slot:
                            merged = defaultdict(dict)
                            for g in slot.get('gains', []):
                                c_id = int(g['id'])
                                if c_id in self.__beq_channels:
                                    merged[c_id]['gain'] = g['value']
                                else:
                                    raise ValueError(f'Invalid channel id for gain setting {c_id}')
                            for g in slot.get('mutes', []):
                                c_id = int(g['id'])
                                if c_id in self.__beq_channels:
                                    merged[c_id]['mute'] = g['value']
                                else:
                                    raise ValueError(f'Invalid channel id for gain setting {c_id}')

                            def completed(success: bool):
                                logger.log(logging.INFO if success else logging.WARNING,
                                           f'Completed gain update {success}')
                                if success:
                                    css = self._current_state.slot
                                    for k, v in merged.items():
                                        if 'gain' in v:
                                            to_update = next(g for g in css.gains if g['id'] == k)
                                            to_update['value'] = v['gain']
                                        if 'mute' in v:
                                            to_update = next(g for g in css.mutes if g['id'] == k)
                                            to_update['value'] = v['mute']

                            self.__update_channel_levels(merged,
                                                         lambda b: self._hydrate_cache_broadcast(lambda: completed(b)))
                            any_update = True
            if 'mute' in params and params['mute'] != self._current_state.mute:
                if self._current_state.mute:
                    self.unmute(None, None)
                else:
                    self.mute(None, None)
                any_update = True
            if 'masterVolume' in params and not math.isclose(params['masterVolume'], self._current_state.master_volume):
                self.set_gain(None, None, params['masterVolume'])
                any_update = True
            return any_update

        return self._hydrate_cache_broadcast(__do_it)

    def __update_channel_levels(self, values: Dict[int, dict], on_complete: Callable[[bool], None]):
        if self.__config_updater is None:
            logger.info(f"Sending {len(values)} level changes")
            self.__config_updater = UpdateGain(values, self.ws_server, self.__ws_client, on_complete)
            self.__config_updater.send_get_config()
        else:
            raise ValueError(f'Unable to load BEQ, config update already in progress for {self.__config_updater.name}')

    def __send_filter(self, entry: Optional[CatalogueEntry], mv_adjust: float, on_complete: Callable[[bool], None]):
        if self.__config_updater is None:
            if entry:
                logger.info(f"Sending {len(entry.filters)} filters for {entry.formatted_title}")
            else:
                logger.info(f"Clearing filters")
            self.__config_updater = LoadConfig(entry, self.__beq_channels, mv_adjust, self.ws_server, self.__ws_client,
                                               on_complete)
            self.__config_updater.send_get_config()
        else:
            raise ValueError(f'Unable to load BEQ, config update already in progress for {self.__config_updater.name}')

    def __send_command(self, set_cmd: str, value: Union[float, bool]):
        logger.info(f"Sending command {set_cmd}: {value}")
        # messages are guaranteed to be processed in order
        self.__ws_client.send(json.dumps({f'Set{set_cmd}': value}))
        self.__ws_client.send(json.dumps(f'Get{set_cmd}'))

    def activate(self, slot: str) -> None:
        def __do_it():
            self._current_state.slot.active = True

        self._hydrate_cache_broadcast(__do_it)

    def load_biquads(self, slot: str, overwrite: bool, inputs: List[int], outputs: List[int],
                     biquads: List[dict]) -> None:
        raise NotImplementedError()

    def send_commands(self, slot: str, inputs: List[int], outputs: List[int], commands: List[str]) -> None:
        raise NotImplementedError()

    def load_filter(self, slot: str, entry: CatalogueEntry, mv_adjust: float = 0.0) -> None:
        self._hydrate_cache_broadcast(lambda: self.__do_load_filter(entry, mv_adjust))

    def __do_load_filter(self, entry: Optional[CatalogueEntry], mv_adjust: float = 0.0):
        try:
            self._current_state.slot.last = 'Loading' if entry else 'Clearing'
            self._current_state.slot.last_author = None

            def completed(success: bool):
                if success:
                    if entry:
                        self._current_state.slot.last = entry.formatted_title
                        self._current_state.slot.last_author = entry.author
                    else:
                        self._current_state.slot.last = 'Empty'
                        self._current_state.slot.last_author = None
                    for g in self._current_state.slot.gains:
                        g['value'] = mv_adjust
                    for g in self._current_state.slot.mutes:
                        g['value'] = False
                else:
                    self._current_state.slot.last = 'ERROR'
                    self._current_state.slot.last_author = None

            self.__send_filter(entry, mv_adjust, lambda b: self._hydrate_cache_broadcast(lambda: completed(b)))
        except Exception as e:
            self._current_state.slot.last = 'ERROR'
            raise e

    def clear_filter(self, slot: str) -> None:
        self._hydrate_cache_broadcast(lambda: self.__do_load_filter(None))

    def mute(self, slot: Optional[str], channel: Optional[int]) -> None:
        self._hydrate_cache_broadcast(lambda: self.__do_mute_op(True))

    def __do_mute_op(self, mute: bool):
        try:
            self.__send_command('Mute', mute)
        except Exception as e:
            raise e

    def unmute(self, slot: Optional[str], channel: Optional[int]) -> None:
        self._hydrate_cache_broadcast(lambda: self.__do_mute_op(False))

    def set_gain(self, slot: Optional[str], channel: Optional[int], gain: float) -> None:
        if channel is None:
            self._hydrate_cache_broadcast(lambda: self.__do_volume_op(gain))
        else:
            self._hydrate_cache_broadcast(lambda: self.__do_gain_op(channel, gain))

    def __do_gain_op(self, channel: int, gain: float):
        pass

    def __do_volume_op(self, level: float):
        try:
            self.__send_command('Volume', level)
        except Exception as e:
            raise e

    def request_levels(self):
        self.__ws_client.send(json.dumps('GetPlaybackSignalPeak'))
        self.__ws_client.send(json.dumps('GetPlaybackSignalRms'))

    def levels(self) -> dict:
        return {
            'name': self.name,
            'ts': time.time(),
            'levels': self.__format_levels()
        }

    def __format_levels(self):
        return {
            **{f'RMS-{i}': v for i, v in enumerate(self.__playback_rms)},
            **{f'Peak-{i}': v for i, v in enumerate(self.__playback_peak)}
        }

    def on_open(self):
        self.__ws_client.send(json.dumps('GetConfigJson'))
        self.__ws_client.send(json.dumps('GetVolume'))
        self.__ws_client.send(json.dumps('GetMute'))
        self.__ws_client.send(json.dumps({'SetUpdateInterval': self.__levels_interval_ms}))

    def on_get_config(self, config: dict):
        if config.get('result', None) == 'Ok':
            new_config = json.loads(config['value'])
            self.current_config = new_config
            if self.__config_updater is None:
                logger.info(f'Received new DSP config but nothing to load, ignoring {new_config}')
            else:
                self.__config_updater.on_get_config(new_config)
        else:
            if self.__config_updater is not None:
                self.__config_updater.failed('GetConfig', config)
                self.__config_updater = None
            else:
                logger.warning(f'GetConfig failed :: {config}')

    def on_set_config(self, result: str):
        if self.__config_updater is None:
            logger.info(f'Received response to SetConfigJson but nothing to load, ignoring {result}')
        else:
            if result == 'Ok':
                self.__config_updater.on_set_config()
            else:
                self.__config_updater.failed('SetConfig', result)
                self.__config_updater = None

    def on_reload(self, result: str):
        if self.__config_updater is None:
            logger.info(f'Received response to Reload but nothing to load, ignoring {result}')
        else:
            if result == 'Ok':
                self.__config_updater.on_reload()
            else:
                self.__config_updater.failed('Reload', result)
            self.__config_updater = None

    def on_get_volume(self, msg):
        if msg['result'] == 'Ok':
            def do_it():
                self._current_state.master_volume = msg['value']

            self._hydrate_cache_broadcast(do_it)

    def on_set_volume(self, msg):
        if msg['result'] == 'Ok':
            self.__ws_client.send(json.dumps('GetVolume'))
        else:
            # TODO send to UI
            logger.warning(f'Unsuccessful command {msg}')

    def on_set_mute(self, msg):
        if msg['result'] == 'Ok':
            self.__ws_client.send(json.dumps('GetMute'))
        else:
            # TODO send to UI
            logger.warning(f'Unsuccessful command {msg}')

    def on_get_mute(self, msg):
        if msg['result'] == 'Ok':
            def do_it():
                self._current_state.mute = msg['value']

            self._hydrate_cache_broadcast(do_it)

    def on_get_playback_rms(self, msg):
        if msg['result'] == 'Ok':
            self.__playback_rms = [v if not math.isclose(v, -1000.0) else -144.0 for v in msg['value']]
            self.ws_server.levels(self.name, self.levels())

    def on_get_playback_peak(self, msg):
        if msg['result'] == 'Ok':
            self.__playback_peak = [v if not math.isclose(v, -1000.0) else -144.0 for v in msg['value']]


class CamillaDspClient:

    def __init__(self, ip: str, port: int, listener: CamillaDsp):
        self.__factory = CamillaDspClientFactory(listener, f"ws://{ip}:{port}")
        self.__connector = connectWS(self.__factory)

    def send(self, msg: str):
        self.__factory.broadcast(msg)


class CamillaDspProtocol(WebSocketClientProtocol):

    # do_send = None

    def onConnecting(self, transport_details):
        logger.info(f"Connecting to {transport_details}")

    def onConnect(self, response):
        logger.info(f"Connected to {response.peer}")

    def onOpen(self):
        logger.info("Connected to CAMILLADSP")
        self.factory.register(self)
        self.factory.listener.on_open()

    def onClose(self, was_clean, code, reason):
        # self.do_send = None
        if was_clean:
            logger.info(f"Disconnected code: {code} reason: {reason}")
        else:
            logger.warning(f"UNCLEAN! Disconnected code: {code} reason: {reason}")

    def onMessage(self, payload, is_binary):
        if is_binary:
            logger.warning(f"Received {len(payload)} bytes in binary payload, ignoring")
        else:
            try:
                msg: dict = json.loads(payload.decode('utf8'))
                logger.debug(f'>>> {msg}')
                if 'GetConfigJson' in msg:
                    self.factory.listener.on_get_config(msg['GetConfigJson'])
                elif 'SetConfigJson' in msg:
                    self.factory.listener.on_set_config(msg['SetConfigJson']['result'])
                elif 'Reload' in msg:
                    self.factory.listener.on_reload(msg['Reload']['result'])
                elif 'GetVolume' in msg:
                    self.factory.listener.on_get_volume(msg['GetVolume'])
                elif 'GetMute' in msg:
                    self.factory.listener.on_get_mute(msg['GetMute'])
                elif 'GetPlaybackSignalRms' in msg:
                    self.factory.listener.on_get_playback_rms(msg['GetPlaybackSignalRms'])
                elif 'GetPlaybackSignalPeak' in msg:
                    self.factory.listener.on_get_playback_peak(msg['GetPlaybackSignalPeak'])
            except:
                logger.exception(f'Unable to decode {len(payload)} bytes in text payload')


class CamillaDspClientFactory(WebSocketClientFactory, ReconnectingClientFactory):
    protocol = CamillaDspProtocol
    maxDelay = 5
    initialDelay = 0.5

    def __init__(self, listener: CamillaDsp, *args, **kwargs):
        super(CamillaDspClientFactory, self).__init__(*args, **kwargs)
        self.__clients: List[CamillaDspProtocol] = []
        self.listener: CamillaDsp = listener
        self.setProtocolOptions(version=13)

    def clientConnectionFailed(self, connector, reason):
        logger.warning(f"Client connection failed {reason} .. retrying ..")
        super().clientConnectionFailed(connector, reason)

    def clientConnectionLost(self, connector, reason):
        logger.warning(f"Client connection failed {reason} .. retrying ..")
        super().clientConnectionLost(connector, reason)

    def register(self, client: CamillaDspProtocol):
        if client not in self.__clients:
            logger.info(f"Registered device {client.peer}")
            self.__clients.append(client)
        else:
            logger.info(f"Ignoring duplicate device {client.peer}")

    def unregister(self, client: CamillaDspProtocol):
        if client in self.__clients:
            logger.info(f"Unregistering device {client.peer}")
            self.__clients.remove(client)
        else:
            logger.info(f"Ignoring unregistered device {client.peer}")

    def broadcast(self, msg):
        if self.__clients:
            disconnected_clients = []
            for c in self.__clients:
                logger.info(f"Sending to {c.peer} - {msg}")
                try:
                    c.sendMessage(msg.encode('utf8'), isBinary=False)
                except Disconnected as e:
                    logger.exception(f"Failed to send to {c.peer}, discarding")
                    disconnected_clients.append(c)
            for c in disconnected_clients:
                self.unregister(c)
        else:
            raise ValueError(f"No devices connected, ignoring {msg}")


def create_cfg_for_entry(entry: Optional[CatalogueEntry], base_cfg: dict, beq_channels: List[int], mv_adjust: float,
                         mute: bool) -> dict:
    from copy import deepcopy
    new_cfg = deepcopy(base_cfg)
    beq_filters = entry.filters if entry else []
    new_filters = new_cfg.get('filters', None)
    new_filters = new_filters if new_filters else {}
    filters = {k: v for k, v in new_filters.items() if not k.startswith('BEQ_')}
    new_cfg['filters'] = filters
    filter_names = []
    gain_filter_names = {}
    if entry or not math.isclose(mv_adjust, 0.0) or mute is True:
        for c in beq_channels:
            name = f'BEQ_Gain_{c}'
            gain_filter_names[c] = name
            filters[name] = {
                'type': 'Gain',
                'parameters': {
                    'gain': mv_adjust,
                    'inverted': False,
                    'mute': mute
                }
            }
    for i, peq in enumerate(beq_filters):
        name = f'BEQ_{i}_{entry.digest}'
        filters[name] = {
            'type': 'Biquad',
            'parameters': {
                'type': get_filter_type(peq),
                'freq': peq['freq'],
                'q': peq['q'],
                'gain': peq['gain']
            }
        }
        filter_names.append(name)
    if 'pipeline' in new_cfg and new_cfg['pipeline']:
        pipeline = new_cfg['pipeline']
        for channel in beq_channels:
            empty_filter = {'type': 'Filter', 'channel': channel, 'names': []}
            existing = None
            for f in pipeline:
                if f['type'] == 'Filter' and f['channel'] == channel:
                    existing = f
            if existing is None:
                existing = empty_filter
                pipeline.append(existing)
            import re
            new_names = [n for n in existing['names'] if re.match(BEQ_FILTER_NAME_PATTERN, n) is None]
            if channel in gain_filter_names:
                new_names.append(gain_filter_names[channel])
            new_names.extend(filter_names)
            existing['names'] = new_names
    else:
        raise ValueError(f'Unable to load BEQ, dsp config has no pipeline declared')
    return new_cfg


def get_filter_type(peq):
    return 'Lowshelf' if peq['type'] == 'LowShelf' else 'Highshelf' if peq['type'] == 'HighShelf' else 'Peaking'


def create_cfg_for_gains(values: Dict[int, dict], base_cfg: dict) -> dict:
    from copy import deepcopy
    new_cfg = deepcopy(base_cfg)
    if 'filters' not in new_cfg:
        new_cfg['filters'] = {}
    for ch, v in values.items():
        gain_filter_name = f'BEQ_Gain_{ch}'
        gain_filter = new_cfg['filters'].get(gain_filter_name, None)
        if gain_filter is None:
            gain_filter = {
                'type': 'Gain',
                'parameters': {
                    'gain': v.get('gain', 0.0),
                    'inverted': False,
                    'mute': v.get('mute', False)
                }
            }
            new_cfg['filters'][gain_filter_name] = gain_filter
        else:
            if 'gain' in v:
                gain_filter['parameters']['gain'] = v['gain']
            if 'mute' in v:
                gain_filter['parameters']['mute'] = v['mute']

        if 'pipeline' in new_cfg:
            pipeline = new_cfg['pipeline']
            empty_filter = {'type': 'Filter', 'channel': ch, 'names': []}
            existing = next((f for f in pipeline if f['type'] == 'Filter' and f['channel'] == ch), None)
            if existing is None:
                existing = empty_filter
                pipeline.append(existing)
            import re
            if gain_filter_name not in existing['names']:
                insert_at = next((i for i, n in enumerate(existing['names']) if re.match(BEQ_FILTER_NAME_PATTERN, n) is not None), -1)
                if insert_at == -1:
                    existing['names'].append(gain_filter_name)
                else:
                    existing['names'].insert(insert_at, gain_filter_name)
        else:
            raise ValueError(f'Unable to load BEQ, dsp config has no pipeline declared')
    return new_cfg


class UpdateConfig:

    def __init__(self, name: str, create_cfg: Callable[[dict], dict], ws_server: WsServer, client: CamillaDspClient,
                 on_complete: Callable[[bool], None]):
        self.__name = name
        self.__create_cfg = create_cfg
        self.__dsp_config = None
        self.__ws_server = ws_server
        self.__client = client
        self.__failed = False
        self.__on_complete = on_complete

    @property
    def name(self):
        return self.__name

    def on_get_config(self, cfg: dict):
        logger.info(f"[{self.__name}] Received new DSP config {cfg}")
        self.__dsp_config = cfg
        self.__do_set_config()

    def send_get_config(self):
        logger.info(f'[{self.__name}] Sending GetConfigJson')
        self.__client.send(json.dumps("GetConfigJson"))

    def __do_set_config(self):
        logger.info(f'[{self.__name}] Sending SetConfigJson')
        self.__client.send(json.dumps({'SetConfigJson': json.dumps(self.__create_cfg(self.__dsp_config))}))

    def on_set_config(self):
        logger.info(f'[{self.__name}] Sending Reload')
        self.__client.send(json.dumps('Reload'))

    def on_reload(self):
        logger.info(f'[{self.__name}] Reload completed')
        self.__on_complete(True)

    def failed(self, stage: str, payload):
        self.__failed = True
        msg = f'{stage} failed : {payload}'
        logger.warning(f'[{self.__name}] Operation failed - {msg}')
        self.__on_complete(False)
        self.__ws_server.broadcast(json.dumps({'message': 'Error', 'data': msg}))


class LoadConfig(UpdateConfig):

    def __init__(self, entry: Optional[CatalogueEntry], beq_channels: List[int], mv_adjust: float, ws_server: WsServer,
                 client: CamillaDspClient, on_complete: Callable[[bool], None]):
        super().__init__(entry.formatted_title if entry else 'Clearing BEQ',
                         lambda base: create_cfg_for_entry(entry, base, beq_channels, mv_adjust, False),
                         ws_server, client, on_complete)


class UpdateGain(UpdateConfig):
    def __init__(self, values: Dict[int, dict], ws_server: WsServer, client: CamillaDspClient,
                 on_complete: Callable[[bool], None]):
        super().__init__('UpdateGain', lambda base: create_cfg_for_gains(values, base), ws_server, client, on_complete)



import json
import logging
import math
import os
import time
from concurrent.futures.thread import ThreadPoolExecutor
from contextlib import contextmanager
from typing import List, Optional, Union

import yaml
from autobahn.exception import Disconnected
from autobahn.twisted import WebSocketClientFactory, WebSocketClientProtocol
from plumbum import ProcessExecutionError
from twisted.internet.protocol import ReconnectingClientFactory

from ezbeq.apis.ws import WsServer
from ezbeq.catalogue import CatalogueEntry, CatalogueProvider
from ezbeq.device import InvalidRequestError, SlotState, PersistentDevice, DeviceState, UnableToPatchDeviceError
from ezbeq import to_millis

INPUT_NAME = 'input'
OUTPUT_NAME = 'output'
CROSSOVER_NAME = 'crossover'

logger = logging.getLogger('ezbeq.minidsp')


class MinidspState(DeviceState):

    def __init__(self, name: str, descriptor: 'MinidspDescriptor', **kwargs):
        self.__name = name
        self.master_volume: float = kwargs['mv'] if 'mv' in kwargs else 0.0
        self.__mute: bool = kwargs['mute'] if 'mute' in kwargs else False
        self.__active_slot: str = kwargs['active_slot'] if 'active_slot' in kwargs else ''
        self.__descriptor = descriptor
        slot_ids = [str(i + 1) for i in range(4)]
        self.__slots: List[MinidspSlotState] = [
            MinidspSlotState(c_id,
                             c_id == self.active_slot,
                             0 if not descriptor.input else len(descriptor.input.channels),
                             0 if not descriptor.output else len(descriptor.output.channels),
                             slot_name=descriptor.slot_names.get(c_id, None))
            for c_id in slot_ids
        ]

    def update_master_state(self, mute: bool, gain: float):
        self.__mute = mute
        self.master_volume = gain

    def activate(self, slot_id: str):
        self.__active_slot = slot_id
        for s in self.__slots:
            s.active = s.slot_id == slot_id

    @property
    def active_slot(self) -> str:
        return self.__active_slot

    @property
    def mute(self) -> bool:
        return self.__mute

    def load(self, slot_id: str, title: str, author: str = None):
        slot = self.get_slot(slot_id)
        slot.last = title
        slot.last_author = author
        self.activate(slot_id)

    def get_slot(self, slot_id) -> 'MinidspSlotState':
        return next(s for s in self.__slots if s.slot_id == slot_id)

    def clear(self, slot_id):
        slot = self.get_slot(slot_id)
        slot.unmute(None)
        slot.set_gain(None, 0.0)
        slot.last = 'Empty'
        slot.last_author = None
        self.activate(slot_id)

    def error(self, slot_id):
        slot = self.get_slot(slot_id)
        slot.last = 'ERROR'
        slot.last_author = None
        self.activate(slot_id)

    def gain(self, slot_id: Optional[str], channel: Optional[int], gain: float):
        if slot_id is None:
            self.master_volume = gain
        else:
            self.get_slot(slot_id).set_gain(channel, gain)
            self.activate(slot_id)

    def toggle_mute(self, slot_id: Optional[str], channel: Optional[int], mute: bool):
        if slot_id is None:
            self.__mute = mute
        else:
            slot = self.get_slot(slot_id)
            if mute:
                slot.mute(channel)
            else:
                slot.unmute(channel)
            self.activate(slot_id)

    def serialise(self) -> dict:
        return {
            'type': 'minidsp',
            'name': self.__name,
            'masterVolume': self.master_volume,
            'mute': self.__mute,
            'slots': [s.as_dict() for s in self.__slots]
        }

    def merge_with(self, cached: dict) -> None:
        saved_slots_by_id = {v['id']: v for v in cached.get('slots', [])}
        current_slots_by_id = {s.slot_id: s for s in self.__slots}
        if saved_slots_by_id.keys() == current_slots_by_id.keys():
            for slot_id, state in saved_slots_by_id.items():
                current_slots_by_id[slot_id].merge_with(state)


class MinidspSlotState(SlotState['MinidspSlotState']):

    def __init__(self, slot_id: str, active: bool, input_channels: int, output_channels: int, slot_name: str = None):
        super().__init__(slot_id)
        self.__input_channels = input_channels
        self.__output_channels = output_channels
        self.gains = self.__make_vals(0.0)
        self.mutes = self.__make_vals(False)
        self.active = active
        self.slot_name = slot_name

    def clear(self):
        super().clear()
        self.gains = self.__make_vals(0.0)
        self.mutes = self.__make_vals(False)

    def __make_vals(self, val: Union[float, bool]) -> List[dict]:
        return [{'id': str(i + 1), 'value': val} for i in range(self.__input_channels)]

    def set_gain(self, channel: Optional[int], value: float):
        if channel is None:
            self.gains = self.__make_vals(value)
        else:
            if channel <= self.__input_channels:
                next(g for g in self.gains if g['id'] == str(channel))['value'] = value
            else:
                raise ValueError(f'Unknown channel {channel} for slot {self.slot_id}')

    def mute(self, channel: Optional[int]):
        self.__do_mute(channel, True)

    def __do_mute(self, channel: Optional[int], value: bool):
        if channel is None:
            self.mutes = self.__make_vals(value)
        else:
            if channel <= self.__input_channels:
                next(g for g in self.mutes if g['id'] == str(channel))['value'] = value
            else:
                raise ValueError(f'Unknown channel {channel} for slot {self.slot_id}')

    def unmute(self, channel: Optional[int]):
        self.__do_mute(channel, False)

    def merge_with(self, state: dict) -> None:
        super().merge_with(state)
        if 'gains' in state and len(state['gains']) == self.__input_channels:
            self.gains = []
            for i, g in enumerate(state['gains']):
                if isinstance(g, dict):
                    self.gains.append(g)
                else:
                    self.gains.append({'id': str(i+1), 'value': float(g)})
        if 'mutes' in state and len(state['mutes']) == self.__input_channels:
            self.mutes = []
            for i, m in enumerate(state['mutes']):
                if isinstance(m, dict):
                    self.mutes.append(m)
                else:
                    self.mutes.append({'id': str(i+1), 'value': bool(m)})

    def as_dict(self) -> dict:
        sup = super().as_dict()
        if self.slot_name:
            sup['name'] = self.slot_name
        return {
            **sup,
            'gains': self.gains,
            'mutes': self.mutes,
            'canActivate': True,
            'inputs': self.__input_channels,
            'outputs': self.__output_channels,
        }

    def __repr__(self):
        vals = ' '.join([f"{g['id']}: {g['value']:.2f}/{self.mutes[i]['value']}" for i, g in enumerate(self.gains)])
        return f"{super().__repr__()} - {vals}"


class PeqRoutes:
    def __init__(self, name: str, biquads: int, channels: List[int], beq_slots: List[int], groups: List[int] = None):
        self.name = name
        self.biquads = biquads
        self.channels = channels
        self.beq_slots = beq_slots
        self.groups = groups

    @property
    def takes_beq(self) -> bool:
        return self.channels and len(self.channels) > 0 and self.beq_slots and len(self.beq_slots) > 0

    def __repr__(self):
        return f"{self.name}"

    def __eq__(self, o: object) -> bool:
        if isinstance(o, PeqRoutes):
            same = self.name == o.name and self.biquads == o.biquads and self.channels == o.channels and self.beq_slots == o.beq_slots
            if same:
                return (self.groups is None and o.groups is None) or (
                        self.groups is not None and self.groups == o.groups)
            return same
        return NotImplemented


class BeqFilterSlot:

    def __init__(self, name: str, idx: int, channels: List[int], group: Optional[int] = None):
        self.name = name
        self.idx = idx
        self.channels = channels
        self.group = group

    def __repr__(self):
        return f"{self.name}{self.group if self.group is not None else ''}/{self.idx}/{self.channels}"


class BeqFilterAllocator:

    def __init__(self, routes: List[PeqRoutes]):
        self.slots = []
        for r in routes:
            if r and r.takes_beq:
                for s in r.beq_slots:
                    if r.groups:
                        for g in r.groups:
                            self.slots.append(BeqFilterSlot(r.name, s, r.channels, g))
                    else:
                        self.slots.append(BeqFilterSlot(r.name, s, r.channels))

    def pop(self) -> Optional[BeqFilterSlot]:
        if self.slots:
            return self.slots.pop(0)
        return None

    def __len__(self):
        return len(self.slots)

    def __repr__(self):
        return f"{self.slots}"


class MinidspDescriptor:

    def __init__(self, name: str, fs: str, i: Optional[PeqRoutes] = None, xo: Optional[PeqRoutes] = None,
                 o: Optional[PeqRoutes] = None, extra: List[PeqRoutes] = None, slot_names: dict[str, str] = None):
        self.name = name
        self.fs = str(int(fs))
        self.input = i
        self.crossover = xo
        self.output = o
        self.extra = extra
        self.slot_names = slot_names if slot_names else {}

    @property
    def peq_routes(self) -> List[PeqRoutes]:
        return [x for x in [self.input, self.crossover, self.output] if x] + (
            [x for x in self.extra] if self.extra else [])

    def to_allocator(self) -> BeqFilterAllocator:
        return BeqFilterAllocator(self.peq_routes)

    def __repr__(self):
        s = f"{self.name}, fs:{self.fs}"
        if self.input:
            s = f"{s}, inputs: {self.input}"
        if self.crossover:
            s = f"{s}, crossovers: {self.crossover}"
        if self.output:
            s = f"{s}, outputs: {self.output}"
        if self.slot_names:
            s = f"{s}, slot_names: {self.slot_names}"
        return s


def zero_til(count: int) -> List[int]:
    return list(range(0, count))


class Minidsp24HD(MinidspDescriptor):

    def __init__(self, slot_names: dict[str, str] = None):
        super().__init__('2x4HD',
                         '96000',
                         i=PeqRoutes(INPUT_NAME, 10, zero_til(2), zero_til(10)),
                         xo=PeqRoutes(CROSSOVER_NAME, 4, zero_til(4), [], groups=zero_til(2)),
                         o=PeqRoutes(OUTPUT_NAME, 10, zero_til(4), []),
                         slot_names=slot_names)


class MinidspDDRC24(MinidspDescriptor):

    def __init__(self, slot_names: dict[str, str] = None):
        super().__init__('DDRC24',
                         '48000',
                         xo=PeqRoutes(CROSSOVER_NAME, 4, zero_til(4), [], zero_til(2)),
                         o=PeqRoutes(OUTPUT_NAME, 10, zero_til(4), zero_til(10)))


class MinidspDDRC88(MinidspDescriptor):

    def __init__(self, slot_names: dict[str, str] = None, sw_channels: List[int] = None):
        c = sw_channels if sw_channels is not None else [3]
        if any(ch for ch in c if ch < 0 or ch > 7):
            raise ValueError(f"Invalid channels {c}")
        non_sw = [c1 for c1 in zero_til(8) if c1 not in c]
        super().__init__('DDRC88',
                         '48000',
                         xo=PeqRoutes(CROSSOVER_NAME, 8, zero_til(8), [], zero_til(2)),
                         o=PeqRoutes(OUTPUT_NAME, 10, c, zero_til(10)),
                         extra=[PeqRoutes(OUTPUT_NAME, 10, non_sw, []) if non_sw else None])


class Minidsp410(MinidspDescriptor):

    def __init__(self, slot_names: dict[str, str] = None):
        super().__init__('4x10',
                         '96000',
                         i=PeqRoutes(INPUT_NAME, 5, zero_til(2), zero_til(5)),
                         o=PeqRoutes(OUTPUT_NAME, 5, zero_til(8), zero_til(5)))


class Minidsp1010(MinidspDescriptor):

    def __init__(self, use_xo: Union[bool, int, str], slot_names: dict[str, str] = None):
        if use_xo is True:
            secondary = {'xo': PeqRoutes(CROSSOVER_NAME, 4, zero_til(8), zero_til(4), groups=[0])}
        elif use_xo is False:
            secondary = {'o': PeqRoutes(OUTPUT_NAME, 6, zero_til(8), zero_til(4))}
        elif use_xo == '0' or use_xo == '1':
            secondary = {'xo': PeqRoutes(CROSSOVER_NAME, 4, zero_til(8), zero_til(4), groups=[int(use_xo)])}
        elif use_xo == 0 or use_xo == 1:
            secondary = {'xo': PeqRoutes(CROSSOVER_NAME, 4, zero_til(8), zero_til(4), groups=[use_xo])}
        elif use_xo == 'all':
            secondary = {'xo': PeqRoutes(CROSSOVER_NAME, 4, zero_til(8), zero_til(4), groups=zero_til(2))}
        else:
            secondary = {'o': PeqRoutes(OUTPUT_NAME, 6, zero_til(8), zero_til(4))}
        super().__init__('10x10',
                         '48000',
                         i=PeqRoutes(INPUT_NAME, 6, zero_til(8), zero_til(6)),
                         **secondary)


def make_peq_layout(cfg: dict) -> MinidspDescriptor:
    slot_names: dict[str, str] = {str(k): str(v) for k,v in cfg.get('slotNames', {}).items()}
    if 'device_type' in cfg:
        device_type = cfg['device_type']
        if device_type == '24HD':
            return Minidsp24HD(slot_names=slot_names)
        elif device_type == 'DDRC24':
            return MinidspDDRC24(slot_names=slot_names)
        elif device_type == 'DDRC88':
            return MinidspDDRC88(sw_channels=cfg.get('sw_channels', None), slot_names=slot_names)
        elif device_type == '4x10':
            return Minidsp410(slot_names=slot_names)
        elif device_type == '10x10':
            return Minidsp1010(cfg.get('use_xo', False), slot_names=slot_names)
        elif device_type == 'SHD':
            return MinidspDDRC24(slot_names=slot_names)
    elif 'descriptor' in cfg:
        desc: dict = cfg['descriptor']
        named_args = ['name', 'fs', 'routes']
        missing_keys = [x for x in named_args if x not in desc.keys()]
        if missing_keys:
            raise ValueError(f"Custom descriptor is missing keys - {missing_keys} - from {desc}")
        routes: List[dict] = desc['routes']

        def make_route(r: dict) -> PeqRoutes:
            r_named_args = ['name', 'biquads', 'channels', 'slots']
            missing_route_keys = [x for x in r_named_args if x not in r.keys()]
            if missing_route_keys:
                raise ValueError(f"Custom PeqRoutes is missing keys - {missing_route_keys} - from {r}")

            def to_ints(v):
                return [int(i) for i in v] if v else None

            return PeqRoutes(r['name'], int(r['biquads']), to_ints(r['channels']), to_ints(r['slots']),
                             to_ints(r.get('groups', None)))

        routes_by_name = {}
        extra = []
        for r in routes:
            route = make_route(r)
            if route.name == 'input':
                if 'i' in routes_by_name:
                    extra.append(route)
                else:
                    routes_by_name['i'] = route
            elif route.name == 'output':
                if 'o' in routes_by_name:
                    extra.append(route)
                else:
                    routes_by_name['o'] = route
            elif route.name == 'xo' or route.name == CROSSOVER_NAME:
                if 'xo' in routes_by_name:
                    extra.append(route)
                else:
                    routes_by_name['xo'] = route
            else:
                extra.append(route)
        if extra:
            routes_by_name['extra'] = extra
        return MinidspDescriptor(desc['name'], str(desc['fs']), **routes_by_name, slot_names=slot_names)
    else:
        return Minidsp24HD(slot_names=slot_names)


class Minidsp(PersistentDevice[MinidspState]):

    def __init__(self, name: str, config_path: str, cfg: dict, ws_server: WsServer, catalogue: CatalogueProvider):
        super().__init__(config_path, name, ws_server)
        self.__catalogue = catalogue
        self.__executor = ThreadPoolExecutor(max_workers=1)
        self.__cmd_timeout = cfg.get('cmdTimeout', 10)
        self.__ignore_retcode = cfg.get('ignoreRetcode', False)
        self.__slot_change_delay: Union[bool, int] = cfg.get('slotChangeDelay', False)
        self.__levels_interval = 1.0 / float(cfg.get('levelsFps', 10))
        self.__runner = cfg['make_runner'](cfg['exe'], cfg.get('options', ''))
        ws_device_id = cfg.get('wsDeviceId', None)
        ws_ip = cfg.get('wsIp', '127.0.0.1:5380')
        if ws_device_id is not None and ws_ip:
            self.__ws_client = MinidspRsClient(self, ws_ip, ws_device_id)
        else:
            self.__ws_client = None
        self.__descriptor: MinidspDescriptor = make_peq_layout(cfg)
        logger.info(f"[{name}] Minidsp descriptor is loaded.... exe is {self.__runner}")
        logger.info(yaml.dump(self.__descriptor, indent=2, default_flow_style=False, sort_keys=False))
        ws_server.factory.set_levels_provider(name, self.start_broadcast_levels)

    @property
    def device_type(self) -> str:
        return self.__class__.__name__.lower()

    def __load_state(self) -> MinidspState:
        result = self.__executor.submit(self.__read_state_from_device).result(timeout=self.__cmd_timeout)
        return result if result else MinidspState(self.name, self.__descriptor)

    def __read_state_from_device(self) -> Optional[MinidspState]:
        output = None
        try:
            kwargs = {'retcode': None} if self.__ignore_retcode else {}
            output = self.__runner['-o', 'jsonline'](timeout=self.__cmd_timeout, **kwargs)
            lines = output.splitlines()
            if lines:
                status = json.loads(lines[0])
                values = {
                    'active_slot': str(status['master']['preset'] + 1),
                    'mute': status['master']['mute'],
                    'mv': status['master']['volume']
                }
                return MinidspState(self.name, self.__descriptor, **values)
            else:
                logger.error(f"[{self.name}] No output returned from device")
        except:
            logger.exception(f"[{self.name}] Unable to parse device state {output}")
        return None

    @staticmethod
    def __as_idx(idx: Union[int, str]):
        return int(idx) - 1

    def __send_cmds(self, target_slot_idx: Optional[int], cmds: List[str]):
        return self.__executor.submit(self.__do_run, cmds, target_slot_idx, self.__slot_change_delay).result(
            timeout=self.__cmd_timeout)

    def activate(self, slot: str):
        def __do_it():
            target_slot_idx = self.__as_idx(slot)
            self.__validate_slot_idx(target_slot_idx)
            self.__send_cmds(target_slot_idx, [])
            self._current_state.activate(slot)

        self._hydrate_cache_broadcast(__do_it)

    @staticmethod
    def __validate_slot_idx(target_slot_idx):
        if target_slot_idx < 0 or target_slot_idx > 3:
            raise InvalidRequestError(f"Slot must be in range 1-4")

    def load_biquads(self, slot: str, overwrite: bool, inputs: List[int], outputs: List[int],
                     biquads: List[dict]) -> None:
        def __do_it():
            target_slot_idx = self.__as_idx(slot)
            self.__validate_slot_idx(target_slot_idx)
            cmds = MinidspBeqCommandGenerator.biquads(overwrite, inputs, outputs, biquads)
            try:
                self.__send_cmds(target_slot_idx, cmds)
                if inputs:
                    self._current_state.load(slot, 'CUSTOM')
                else:
                    self._current_state.activate(slot)
            except Exception as e:
                self._current_state.error(slot)
                raise e

        self._hydrate_cache_broadcast(__do_it)

    def send_commands(self, slot: str, inputs: List[int], outputs: List[int], commands: List[str]) -> None:
        def __do_it():
            target_slot_idx = self.__as_idx(slot)
            self.__validate_slot_idx(target_slot_idx)
            cmds = MinidspBeqCommandGenerator.commands(inputs, outputs, commands)
            try:
                self.__send_cmds(target_slot_idx, cmds)
                if inputs:
                    self._current_state.load(slot, 'CUSTOM')
                else:
                    self._current_state.activate(slot)
            except Exception as e:
                self._current_state.error(slot)
                raise e

        self._hydrate_cache_broadcast(__do_it)

    def load_filter(self, slot: str, entry: CatalogueEntry, mv_adjust: float = 0.0) -> None:
        def __do_it():
            target_slot_idx = self.__as_idx(slot)
            self.__validate_slot_idx(target_slot_idx)
            cmds = MinidspBeqCommandGenerator.filt(entry, self.__descriptor)
            try:
                self.__send_cmds(target_slot_idx, cmds)
                self._current_state.load(slot, entry.formatted_title, entry.author)
            except Exception as e:
                self._current_state.error(slot)
                raise e

        self._hydrate_cache_broadcast(__do_it)

    def clear_filter(self, slot: str) -> None:
        def __do_it():
            target_slot_idx = self.__as_idx(slot)
            self.__validate_slot_idx(target_slot_idx)
            cmds = MinidspBeqCommandGenerator.filt(None, self.__descriptor)
            beq_slots = self.__descriptor.to_allocator()
            levels = []
            handled = []
            s = beq_slots.pop()
            while s is not None:
                for c in s.channels:
                    if s.name == INPUT_NAME and c not in handled:
                        levels.extend(MinidspBeqCommandGenerator.mute(False, target_slot_idx, c, side=s.name))
                        levels.extend(MinidspBeqCommandGenerator.gain(0.0, target_slot_idx, c, side=s.name))
                        handled.append(c)
                s = beq_slots.pop()
            if levels:
                cmds.extend(levels)
            try:
                self.__send_cmds(target_slot_idx, cmds)
                self._current_state.clear(slot)
            except Exception as e:
                self._current_state.error(slot)
                raise e

        self._hydrate_cache_broadcast(__do_it)

    def mute(self, slot: Optional[str], channel: Optional[int]) -> None:
        self.__do_mute_op(slot, channel, True)

    def __do_mute_op(self, slot: Optional[str], channel: Optional[int], state: bool):
        def __do_it():
            target_channel_idx, target_slot_idx = self.__as_idxes(channel, slot)
            if target_slot_idx:
                self.__validate_slot_idx(target_slot_idx)
            cmds = MinidspBeqCommandGenerator.mute(state, target_slot_idx, target_channel_idx)
            self.__send_cmds(target_slot_idx, cmds)
            self._current_state.toggle_mute(slot, channel, state)

        self._hydrate_cache_broadcast(__do_it)

    def unmute(self, slot: Optional[str], channel: Optional[int]) -> None:
        self.__do_mute_op(slot, channel, False)

    def set_gain(self, slot: Optional[str], channel: Optional[int], gain: float) -> None:
        def __do_it():
            target_channel_idx, target_slot_idx = self.__as_idxes(channel, slot)
            cmds = MinidspBeqCommandGenerator.gain(gain, target_slot_idx, target_channel_idx)
            self.__send_cmds(target_slot_idx, cmds)
            self._current_state.gain(slot, channel, gain)

        self._hydrate_cache_broadcast(__do_it)

    def __as_idxes(self, channel, slot):
        target_slot_idx = self.__as_idx(slot) if slot else None
        target_channel_idx = self.__as_idx(channel) if channel else None
        return target_channel_idx, target_slot_idx

    def __do_run(self, config_cmds: List[str], slot: Optional[int], slot_change_delay: Union[bool, int]):
        if slot is not None:
            change_slot = True
            current_state = self.__read_state_from_device()
            if current_state and current_state.active_slot == str(slot + 1):
                change_slot = False
            if change_slot is True:
                if slot_change_delay:
                    self.__do_run([], slot, False)
                    if slot_change_delay is not True and slot_change_delay > 0:
                        from time import sleep
                        logger.info(f"[{self.name}] Sleeping for {slot_change_delay} seconds after config slot change")
                        sleep(slot_change_delay)
                else:
                    logger.info(
                        f"[{self.name}] Activating slot {slot}, current is {current_state.active_slot if current_state else 'UNKNOWN'}")
                    config_cmds.insert(0, MinidspBeqCommandGenerator.activate(slot))
        formatted = '\n'.join(config_cmds)
        logger.info(f"\n{formatted}")
        with tmp_file(config_cmds) as file_name:
            kwargs = {'retcode': None} if self.__ignore_retcode else {}
            exe = self.__runner['-f', file_name]
            logger.info(
                f"[{self.name}] Sending {len(config_cmds)} commands to slot {slot} using {exe} {kwargs if kwargs else ''}")
            start = time.time()
            try:
                code, stdout, stderr = exe.run(timeout=self.__cmd_timeout, **kwargs)
            except ProcessExecutionError as e:
                raise UnableToPatchDeviceError(f'minidsp cmd failed due to : {e.stderr}', False) from e
            end = time.time()
            logger.info(
                f"[{self.name}] Sent {len(config_cmds)} commands to slot {slot} in {to_millis(start, end)}ms - result is {code}")

    def _load_initial_state(self) -> MinidspState:
        return self.__load_state()

    def state(self, refresh: bool = False) -> MinidspState:
        if not self._hydrate() or refresh is True:
            new_state = self.__load_state()
            self._current_state.update_master_state(new_state.mute, new_state.master_volume)
        return self._current_state

    def _merge_state(self, loaded: MinidspState, cached: dict) -> MinidspState:
        loaded.merge_with(cached)
        return loaded

    def update(self, params: dict) -> bool:
        def __do_it() -> bool:
            any_update = False
            if 'slots' in params:
                for slot in params['slots']:
                    any_update |= self.__update_slot(slot)
            if 'mute' in params and params['mute'] != self._current_state.mute:
                if self._current_state.mute:
                    self.unmute(None, None)
                else:
                    self.mute(None, None)
                any_update = True
            if 'masterVolume' in params and not math.isclose(params['masterVolume'], self._current_state.master_volume):
                self.set_gain(None, None, params['masterVolume'])
                any_update = True
            return any_update

        return self._hydrate_cache_broadcast(__do_it)

    def __update_slot(self, slot: dict) -> bool:
        any_update = False
        current_slot = self._current_state.get_slot(slot['id'])
        if not current_slot:
            raise UnableToPatchDeviceError(f'Unknown device slot {slot["id"]}', True)
        match: Optional[CatalogueEntry] = None
        if 'entry' in slot and slot['entry']:
            match = self.__catalogue.find(slot['entry'])
            if not match:
                raise UnableToPatchDeviceError(f'Unknown catalogue entry {slot["entry"]}', True)
        if 'gains' in slot:
            for gain in slot['gains']:
                self.set_gain(current_slot.slot_id, int(gain['id']), gain['value'])
                any_update = True
        if 'mutes' in slot:
            for mute in slot['mutes']:
                if mute['value'] is True:
                    self.mute(current_slot.slot_id, int(mute['id']))
                else:
                    self.unmute(current_slot.slot_id, int(mute['id']))
                any_update = True
        if 'entry' in slot:
            if slot['entry']:
                self.load_filter(current_slot.slot_id, match)
                any_update = True
            else:
                self.clear_filter(current_slot.slot_id)
        if 'active' in slot:
            self.activate(current_slot.slot_id)
            any_update = True
        return any_update

    def levels(self) -> dict:
        return self.__executor.submit(self.__read_levels_from_device).result(timeout=self.__cmd_timeout)

    def __read_levels_from_device(self) -> dict:
        lines = None
        try:
            kwargs = {'retcode': None} if self.__ignore_retcode else {}
            start = time.time()
            lines = self.__runner['-o', 'jsonline'](timeout=self.__cmd_timeout, **kwargs)
            end = time.time()
            levels = json.loads(lines)
            ts = time.time()
            logger.info(f"{self.name},readlevels,{ts},{to_millis(start, end)}")
            return {
                'name': self.name,
                'ts': ts,
                'levels': format_levels(levels)
            }
        except:
            logger.exception(f"[{self.name}] Unable to load levels {lines}")
            return {}

    def start_broadcast_levels(self) -> None:
        if self.__ws_client is None:
            from twisted.internet import reactor
            sched = lambda: reactor.callLater(self.__levels_interval, __send)

            def __send():
                if self.ws_server.levels(self.name, self.levels()):
                    sched()

            sched()

    def on_ws_message(self, msg: dict):
        logger.debug(f"[{self.name}] Received {msg}")
        if 'master' in msg:
            master = msg['master']
            if master:
                def do_it():
                    preset = str(master['preset'] + 1)
                    mv = master['volume']
                    mute = master['mute']
                    if self._current_state.master_volume != mv:
                        self._current_state.master_volume = mv
                    if self._current_state.mute != mute:
                        self._current_state.mute = mute
                    if self._current_state.active_slot != preset:
                        self._current_state.activate(preset)

                self._hydrate_cache_broadcast(do_it)
        if 'input_levels' in msg and 'output_levels' in msg:
            self.ws_server.levels(self.name, {
                'name': self.name,
                'ts': time.time(),
                'levels': format_levels(msg)
            })


class MinidspBeqCommandGenerator:

    @staticmethod
    def activate(slot: int) -> str:
        return f"config {slot}"

    @staticmethod
    def biquads(overwrite: bool, inputs: List[int], outputs: List[int], biquads: List[dict]):
        # [in|out]put <channel> peq <index> set -- <b0> <b1> <b2> <a1> <a2>
        # [in|out]put <channel> peq <index> bypass [on|off]
        cmds = []
        for side, channels in {INPUT_NAME: inputs, OUTPUT_NAME: outputs}.items():
            for channel in channels:
                for idx, bq in enumerate(biquads):
                    if bq:
                        coeffs = [bq['b0'], bq['b1'], bq['b2'], bq['a1'], bq['a2']]
                        cmds.append(MinidspBeqCommandGenerator.bq(channel - 1, idx, coeffs, side=side))
                        bypass = 'BYPASS' in bq and bq['BYPASS'] is True
                        cmds.append(MinidspBeqCommandGenerator.bypass(channel - 1, idx, bypass, side=side))
                    elif overwrite:
                        cmds.append(MinidspBeqCommandGenerator.bypass(channel - 1, idx, True, side=side))
                if overwrite:
                    for idx in range(len(biquads), 10):
                        cmds.append(MinidspBeqCommandGenerator.bypass(channel - 1, idx, True, side=side))
        return cmds

    @staticmethod
    def commands(inputs: List[int], outputs: List[int], commands: List[str]):
        cmds = []
        for side, channels in {INPUT_NAME: inputs, OUTPUT_NAME: outputs}.items():
            for channel in channels:
                for command in commands:
                    cmds.append(MinidspBeqCommandGenerator.cmd(channel - 1, command, side=side))
        return cmds

    @staticmethod
    def as_bq(f: dict, fs: str):
        if fs in f['biquads']:
            bq = f['biquads'][fs]['b'] + f['biquads'][fs]['a']
        else:
            t = f['type']
            freq = f['freq']
            gain = f['gain']
            q = f['q']
            from ezbeq.iir import PeakingEQ, LowShelf, HighShelf
            if t == 'PeakingEQ':
                f = PeakingEQ(int(fs), freq, q, gain)
            elif t == 'LowShelf':
                f = LowShelf(int(fs), freq, q, gain)
            elif t == 'HighShelf':
                f = HighShelf(int(fs), freq, q, gain)
            else:
                raise InvalidRequestError(f"Unknown filt_type {t}")
            bq = list(f.format_biquads().values())
        if len(bq) != 5:
            raise ValueError(f"Invalid coeff count {len(bq)}")
        return bq

    @staticmethod
    def filt(entry: Optional[CatalogueEntry], descriptor: MinidspDescriptor):
        # [in|out]put <channel> peq <index> set -- <b0> <b1> <b2> <a1> <a2>
        # [in|out]put <channel> peq <index> bypass [on|off]
        cmds = []
        # write filts to the inputs first then the output if it's a split device
        filters = [MinidspBeqCommandGenerator.as_bq(f, descriptor.fs) for f in entry.filters] if entry else []
        beq_slots = descriptor.to_allocator()

        def push(chs: List[int], i: int, s: str, group: Optional[int]):
            for ch in chs:
                cmds.append(MinidspBeqCommandGenerator.bq(ch, i, coeffs, s, group=group))
                cmds.append(MinidspBeqCommandGenerator.bypass(ch, i, False, s, group=group))

        idx = 0
        while idx < len(filters):
            coeffs: List[str] = filters[idx]
            slot = beq_slots.pop()
            if slot is not None:
                push(slot.channels, slot.idx, slot.name, slot.group)
            else:
                raise ValueError(f"Loaded {idx} filters but no slots remaining")
            idx += 1
        s = beq_slots.pop()
        while s is not None:
            for c in s.channels:
                cmds.append(MinidspBeqCommandGenerator.bypass(c, s.idx, True, s.name, s.group))
            s = beq_slots.pop()
        return cmds

    @staticmethod
    def bq(channel: int, idx: int, coeffs, side: str = INPUT_NAME, group: Optional[int] = None):
        is_xo = side == CROSSOVER_NAME
        addr = f"crossover {group}" if is_xo and group is not None else 'peq'
        return f"{OUTPUT_NAME if is_xo else side} {channel} {addr} {idx} set -- {' '.join(coeffs)}"

    @staticmethod
    def cmd(channel: int, cmd: str, side: str = INPUT_NAME):
        return f"{side} {channel} {cmd}"

    @staticmethod
    def bypass(channel: int, idx: int, bypass: bool, side: str = INPUT_NAME, group: Optional[int] = 0):
        is_xo = side == CROSSOVER_NAME
        addr = f"crossover {group}" if is_xo and group is not None else 'peq'
        return f"{OUTPUT_NAME if is_xo else side} {channel} {addr} {idx} bypass {'on' if bypass else 'off'}"

    @staticmethod
    def mute(state: bool, slot: Optional[int], channel: Optional[int], side: Optional[str] = INPUT_NAME):
        '''
        Generates commands to mute the configuration.
        :param state: mute if true otherwise unmute.
        :param slot: the target slot, if not set apply to the master control.
        :param channel: the channel, applicable only if slot is set, if not set apply to both input channels.
        :param side: the side, input by default.
        :return: the commands.
        '''
        state_cmd = 'on' if state else 'off'
        if slot is not None:
            cmds = []
            if channel is None:
                cmds.append(f"{side} 0 mute {state_cmd}")
                cmds.append(f"{side} 1 mute {state_cmd}")
            else:
                cmds.append(f"{side} {channel} mute {state_cmd}")
            return cmds
        else:
            return [f"mute {state_cmd}"]

    @staticmethod
    def gain(gain: float, slot: Optional[int], channel: Optional[int], side: Optional[str] = INPUT_NAME):
        '''
        Generates commands to set gain.
        :param gain: the gain to set.
        :param slot: the target slot, if not set apply to the master control.
        :param channel: the channel, applicable only if slot is set, if not set apply to both input channels.
        :param side: the side to apply the gain to, input by default.
        :return: the commands.
        '''
        if slot is not None:
            # TODO is this valid for other devices
            if not -72.0 <= gain <= 12.0:
                raise InvalidRequestError(f"{side} gain {gain:.2f} out of range (>= -72.0 and <= 12.0)")
            cmds = []
            if channel is None:
                cmds.append(f"{side} 0 gain -- {gain:.2f}")
                cmds.append(f"{side} 1 gain -- {gain:.2f}")
            else:
                cmds.append(f"{side} {channel} gain -- {gain:.2f}")
            return cmds
        else:
            if not -127.0 <= gain <= 0.0:
                raise InvalidRequestError(f"Master gain {gain:.2f} out of range (>= -127.0 and <= 0.0)")
            return [f"gain -- {gain:.2f}"]


@contextmanager
def tmp_file(cmds: List[str]):
    import tempfile
    tmp_name = None
    try:
        f = tempfile.NamedTemporaryFile(mode='w+', delete=False)
        for cmd in cmds:
            f.write(cmd)
            f.write('\n')
        tmp_name = f.name
        f.close()
        yield tmp_name
    finally:
        if tmp_name:
            os.unlink(tmp_name)


class MinidspRsClient:

    def __init__(self, listener, ip, device_id):
        ws_url = f"ws://{ip}/devices/{device_id}?levels=true&poll=true"
        logger.info(f"Listening to ws on {ws_url}")
        self.__factory = MinidspRsClientFactory(listener, device_id, url=ws_url)
        from twisted.internet.endpoints import clientFromString
        from twisted.internet import reactor
        # wsclient = clientFromString(reactor, 'unix:path=/tmp/minidsp.sock:timeout=5')
        wsclient = clientFromString(reactor, f"tcp:{ip}:timeout=5")
        self.__connector = wsclient.connect(self.__factory)

    def send(self, msg: str):
        self.__factory.broadcast(msg)


class MinidspRsProtocol(WebSocketClientProtocol):

    def onConnecting(self, transport_details):
        logger.info(f"Connecting to {transport_details}")

    def onConnect(self, response):
        logger.info(f"Connected to {response.peer}")

    def onOpen(self):
        logger.info("Connected to Minidsp")
        self.factory.register(self)

    def onClose(self, was_clean, code, reason):
        if was_clean:
            logger.info(f"Disconnected code: {code} reason: {reason}")
        else:
            logger.warning(f"UNCLEAN! Disconnected code: {code} reason: {reason}")

    def onMessage(self, payload, is_binary):
        if is_binary:
            logger.warning(f"Received {len(payload)} bytes in binary payload, ignoring")
        else:
            msg = payload.decode('utf8')
            logger.debug(f"[{self.factory.device_id}] Received {msg}")
            try:
                self.factory.listener.on_ws_message(json.loads(msg))
            except:
                logger.exception(f"[{self.factory.device_id}] Receiving unparseable message {msg}")


class MinidspRsClientFactory(WebSocketClientFactory, ReconnectingClientFactory):
    protocol = MinidspRsProtocol
    maxDelay = 5
    initialDelay = 0.5

    def __init__(self, listener, device_id, *args, **kwargs):
        super(MinidspRsClientFactory, self).__init__(*args, **kwargs)
        self.__device_id = device_id
        self.__clients: List[MinidspRsProtocol] = []
        self.listener = listener

    @property
    def device_id(self):
        return self.__device_id

    def clientConnectionFailed(self, connector, reason):
        logger.warning(f"[{self.device_id}] Client connection failed {reason} .. retrying ..")
        super().clientConnectionFailed(connector, reason)

    def clientConnectionLost(self, connector, reason):
        logger.warning(f"[{self.device_id}] Client connection failed {reason} .. retrying ..")
        super().clientConnectionLost(connector, reason)

    def register(self, client: MinidspRsProtocol):
        if client not in self.__clients:
            logger.info(f"[{self.device_id}] Registered device {client.peer}")
            self.__clients.append(client)
        else:
            logger.info(f"[{self.device_id}] Ignoring duplicate device {client.peer}")

    def unregister(self, client: MinidspRsProtocol):
        if client in self.__clients:
            logger.info(f"Unregistering device {client.peer}")
            self.__clients.remove(client)
        else:
            logger.info(f"Ignoring unregistered device {client.peer}")

    def broadcast(self, msg):
        if self.__clients:
            disconnected_clients = []
            for c in self.__clients:
                logger.info(f"[{self.device_id}] Sending to {c.peer} - {msg}")
                try:
                    c.sendMessage(msg.encode('utf8'))
                except Disconnected as e:
                    logger.exception(f"[{self.device_id}] Failed to send to {c.peer}, discarding")
                    disconnected_clients.append(c)
            for c in disconnected_clients:
                self.unregister(c)
        else:
            raise ValueError(f"No devices connected, ignoring {msg}")


def format_levels(levels: dict) -> dict:
    # quick hack for testing purposes
    # INPUT_NAME: [x + ((random() * 5) * (-1.0 if self.name == 'd1' else 1.0)) for x in msg['input_levels']],
    return {
        **{f'I{i}': v for i, v in enumerate(levels['input_levels'])},
        **{f'O{i}': v for i, v in enumerate(levels['output_levels'])}
    }



import logging
import xml.etree.ElementTree as et
from typing import Optional, List, Dict, Tuple

import requests

from ezbeq.apis.ws import WsServer
from ezbeq.catalogue import CatalogueEntry, CatalogueProvider
from ezbeq.device import SlotState, DeviceState, PersistentDevice

logger = logging.getLogger('ezbeq.jriver')

JRIVER_CHANNELS = [None, None, 'L', 'R', 'C', 'SW', 'SL', 'SR', 'RL', 'RR', None, 'U1', 'U2'] + [f"C{i + 9}" for i in
                                                                                                 range(24)]


class JRiverSlotState(SlotState):

    def __init__(self, zone_id: str, zone_name: str):
        super().__init__(zone_name)
        self.__zone_id = zone_id

    @property
    def zone_id(self) -> str:
        return self.__zone_id


class JRiverState(DeviceState):

    def __init__(self, name: str, zones: Dict[str, dict]):
        self.__name = name
        self.__slots: Dict[str, JRiverSlotState] = {z['name']: JRiverSlotState(z_id, z['name'])
                                                    for z_id, z in zones.items()}

    def has_zone(self, zone: str) -> bool:
        slot = self.__slots.get(zone, None)
        return slot is not None

    def get_zone_id(self, zone: str) -> Optional[str]:
        slot = self.__slots.get(zone, None)
        return slot.zone_id if slot else None

    def set_title(self, zone: str, title: str, ignore: bool = False, author: str = None):
        slot = self.__slots.get(zone, None)
        if slot:
            slot.last = title
            slot.last_author = author
        elif not ignore:
            raise KeyError(f"No zone {zone}")

    def activate(self, zone: str):
        for s in self.__slots.values():
            s.active = s.slot_id == zone

    def error(self, zone: str):
        self.__slots[zone].last = 'ERROR'
        self.__slots[zone].last_author = None

    def serialise(self) -> dict:
        return {
            'type': 'jriver',
            'name': self.__name,
            'slots': [s.as_dict() for s in self.__slots.values()]
        }


class JRiver(PersistentDevice[JRiverState]):

    def __init__(self, name: str, config_path: str, cfg: dict, ws_server: WsServer, catalogue: CatalogueProvider):
        super().__init__(config_path, name, ws_server)
        self.__catalogue = catalogue
        address: str = cfg['address']
        if not address:
            raise ValueError('No MCWS address for jriver')
        cfg_auth = cfg.get('auth', None)
        auth = (cfg_auth['user'], cfg_auth['pass']) if cfg_auth else None
        secure = bool(cfg.get('secure', False))
        self.__channels: str = ';'.join([str(JRIVER_CHANNELS.index(c)) for c in cfg['channels']])
        if not self.__channels:
            raise ValueError('No target channels for jriver')
        self.__peq_block: str = get_peq_key_name(int(cfg['block']) - 1)
        if not self.__peq_block:
            raise ValueError('No peq block for jriver')
        self.__mcws = MediaServer(address, auth, secure)

    def device_type(self) -> str:
        return self.__class__.__name__.lower()

    def _load_initial_state(self) -> JRiverState:
        return JRiverState(self.name, self.__mcws.get_zones())

    def _merge_state(self, loaded: JRiverState, cached: dict) -> JRiverState:
        if 'slots' in cached:
            for slot in cached['slots']:
                if 'id' in slot and 'last' in slot:
                    loaded.set_title(slot['id'], slot['last'], author=slot.get('author', None), ignore=True)
        return loaded

    def activate(self, slot: str) -> None:
        self._hydrate_cache_broadcast(lambda: self._current_state.activate(slot))

    def update(self, params: dict) -> bool:
        any_update = False
        if 'slots' in params:
            for slot in params['slots']:
                if self._current_state.has_zone(slot['id']):
                    if 'entry' in slot:
                        if slot['entry']:
                            match = self.__catalogue.find(slot['entry'])
                            if match:
                                self.load_filter(slot['id'], match)
                                any_update = True
                        else:
                            self.clear_filter(slot['id'])
                            any_update = True
        return any_update

    def load_biquads(self, slot: str, overwrite: bool, inputs: List[int], outputs: List[int], biquads: List[dict]) -> None:
        raise NotImplementedError()

    def send_commands(self, slot: str, inputs: List[int], outputs: List[int], commands: List[str]) -> None:
        raise NotImplementedError()

    def load_filter(self, slot: str, entry: CatalogueEntry, mv_adjust: float = 0.0) -> None:
        def __do_it():
            mc_filts = [make_meta(entry.title, True)] \
                       + [convert_filter_to_mc_dsp(f, self.__channels) for f in entry.filters] \
                       + [make_meta(entry.title, False)]
            xml_filts = [filts_to_xml(f) for f in mc_filts]
            zone_id = self._current_state.get_zone_id(slot)
            current_config_txt = self.__mcws.get_dsp(zone_id)
            new_config_txt = self.__remove_all_beq(current_config_txt)
            new_config_txt = include_filters_in_dsp(self.__peq_block, new_config_txt, xml_filts, replace=False)
            logger.info(new_config_txt)
            try:
                self.__mcws.set_dsp(zone_id, new_config_txt)
                self._current_state.set_title(slot, entry.formatted_title, author=entry.author)
            except Exception as e:
                self._current_state.slot.last = 'ERROR'
                self._current_state.slot.last_author = None
                raise e
        self._hydrate_cache_broadcast(__do_it)

    def __remove_all_beq(self, current_config_txt) -> str:
        tries = 0
        final_config_txt = current_config_txt
        while tries < 100:
            tmp_txt = remove_beq_filter(self.__peq_block, final_config_txt)
            if tmp_txt and tmp_txt != final_config_txt:
                final_config_txt = tmp_txt
            else:
                break
        return final_config_txt

    def clear_filter(self, slot: str) -> None:
        def __do_it():
            zone_id = self._current_state.get_zone_id(slot)
            current_config_txt = self.__mcws.get_dsp(zone_id)
            new_config_txt = self.__remove_all_beq(current_config_txt)
            try:
                if new_config_txt != current_config_txt:
                    self.__mcws.set_dsp(zone_id, new_config_txt)
                self._current_state.set_title(slot, 'Empty')
            except Exception as e:
                self._current_state.error(slot)
                raise e
        self._hydrate_cache_broadcast(__do_it)

    def mute(self, slot: Optional[str], channel: Optional[int]) -> None:
        raise NotImplementedError()

    def unmute(self, slot: Optional[str], channel: Optional[int]) -> None:
        raise NotImplementedError()

    def set_gain(self, slot: Optional[str], channel: Optional[int], gain: float) -> None:
        raise NotImplementedError()

    def levels(self) -> dict:
        # TODO implement
        return {}


def convert_filter_to_mc_dsp(filt: dict, target_channels: str) -> List[Dict[str, str]]:
    '''
    :param filt: filter values in catalogue form.
    :param target_channels: the channels to output to.
    :return: filter values in mc form.
    '''
    f_type = filt['type']
    if f_type == 'LowShelf':
        return make_shelf(target_channels, filt, False)
    elif f_type == 'HighShelf':
        return make_shelf(target_channels, filt, True)
    elif f_type == 'PeakingEQ':
        return [make_peak(target_channels, filt)]
    elif f_type == 'Gain':
        return [make_gain(target_channels, filt)]
    raise ValueError(f"Unknown filter type {f_type} in {filt}")


def make_peak(channels: str, vals: dict) -> Dict[str, str]:
    return {
        'Enabled': '1',
        'Slope': '12',
        'Q': f"{vals['q']}",
        'Type': '3',
        'Gain': f"{vals['gain']}",
        'Frequency': f"{vals['freq']}",
        'Channels': channels
    }


def make_shelf(channels: str, vals: dict, high: bool) -> List[Dict[str, str]]:
    filt_val = {
        'Enabled': '1',
        'Slope': '12',
        'Q': f"{q_to_s(vals['q'], vals['gain']):.12g}",
        'Type': '11' if high else '10',
        'Gain': f"{vals['gain']}",
        'Frequency': f"{vals['freq']}",
        'Channels': channels
    }
    count = int(vals.get('count', 1))
    return [filt_val] * count


def make_gain(channels: str, vals: dict) -> Dict[str, str]:
    return {
        'Enabled': '1',
        'Type': '4',
        'Gain': f"{vals['gain']}",
        'Channels': channels
    }


def make_meta(name: str, start: bool) -> List[Dict[str, str]]:
    suf = 'START' if start else 'END'
    return [{
        'Enabled': '1',
        'Type': '20',
        'Text': f"***BEQ_{suf}|{name}"
    }]


def q_to_s(q: float, gain: float):
    '''
    translates Q to S for a shelf filter.
    :param q: the Q.
    :param gain: the gain.
    :return: the S.
    '''
    return 1.0 / ((((1.0 / q) ** 2.0 - 2.0) / (
            (10.0 ** (gain / 40.0)) + 1.0 / (10.0 ** (gain / 40.0)))) + 1.0)


def xpath_to_key_data_value(key_name, data_name):
    '''
    an ET compatible xpath to get the value from a DSP config via the path /Preset/Key/Data/Value for a given key and
    data.
    :param key_name:
    :param data_name:
    :return:
    '''
    return f"./Preset/Key[@Name=\"{key_name}\"]/Data/Name[.=\"{data_name}\"]/../Value"


class NoFiltersError(ValueError):
    pass


def extract_filters(config_txt: str, key_name: str, allow_empty: bool = False):
    '''
    :param config_txt: the xml text.
    :param key_name: the filter key name.
    :param allow_empty: if true, create the missing filters element if it doesn't exist.
    :return: (root element, filter element)
    '''
    root = et.fromstring(config_txt)
    elements = root.findall(xpath_to_key_data_value(key_name, 'Filters'))
    if elements and len(elements) == 1:
        return root, elements[0]
    if allow_empty:
        parent_element = root.find(f"./Preset/Key[@Name=\"{key_name}\"]")
        data_element = et.Element('Data')
        name_element = et.Element('Name')
        name_element.text = 'Filters'
        data_element.append(name_element)
        value_element = et.Element('Value')
        value_element.text = ''
        data_element.append(value_element)
        parent_element.append(data_element)
        return root, value_element
    else:
        raise NoFiltersError(f"No Filters in {key_name} found in {config_txt}")


def filts_to_xml(vals: List[Dict[str, str]]) -> str:
    '''
    Formats key-value pairs into a jriver dsp config file compatible str fragment.
    :param vals: the key-value pairs.
    :return: the txt snippet.
    '''
    return ''.join(filt_to_xml(f) for f in vals)


def filt_to_xml(vals: Dict[str, str]) -> str:
    '''
    Converts a set of filter values to a jriver compatible xml fragment.
    :param vals: the values.
    :return: the xml fragment.
    '''
    items = [f"<Item Name=\"{k}\">{v}</Item>" for k, v in vals.items()]
    catted_items = '\n'.join(items)
    prefix = '<XMLPH version="1.1">'
    suffix = '</XMLPH>'
    txt_length = len(prefix) + len(''.join(items)) + len(suffix)
    new_line_len = (len(items) + 1) * 2
    total_len = txt_length + new_line_len
    xml_frag = f"({total_len}:{prefix}\n{catted_items}\n{suffix})"
    # print(f"{filter_classes_by_type[vals['Type']].__name__} ({vals['Type']}): {offset}")
    return xml_frag


def include_filters_in_dsp(peq_block_name: str, config_txt: str, xml_filts: List[str], replace: bool = True) -> str:
    '''
    :param peq_block_name: the peq block to process.
    :param config_txt: the dsp config in txt form.
    :param xml_filts: the filters to include.
    :param replace: if true, replace existing filters. if false, append.
    :return: the new config txt.
    '''
    if xml_filts:
        root, filt_element = extract_filters(config_txt, peq_block_name, allow_empty=True)
        # before_value, after_value, filt_section = extract_value_section(config_txt, self.__block)
        # separate the tokens, which are in (TOKEN) blocks, from within the Value element
        if filt_element.text:
            filt_fragments = [v + ')' for v in filt_element.text.split(')') if v]
            if len(filt_fragments) < 2:
                raise ValueError('Invalid input file - Unexpected <Value> format')
        else:
            filt_fragments = ['(1:1)', '(2:0)']
        # find the filter count and replace it with the new filter count
        new_filt_count = sum([x.count('<XMLPH version') for x in xml_filts])
        if not replace:
            new_filt_count = int(filt_fragments[1][1:-1].split(':')[1]) + new_filt_count
        filt_fragments[1] = f"({len(str(new_filt_count))}:{new_filt_count})"
        # append the new filters to any existing ones or replace
        if replace:
            new_filt_section = ''.join(filt_fragments[0:2]) + ''.join(xml_filts)
        else:
            new_filt_section = ''.join(filt_fragments) + ''.join(xml_filts)
        # replace the value block in the original string
        filt_element.text = new_filt_section
        config_txt = et.tostring(root, encoding='UTF-8', xml_declaration=True).decode('utf-8')
        return config_txt
    else:
        return config_txt


def remove_beq_filter(peq_block_name: str, config_txt: str) -> Optional[str]:
    root, filt_element = extract_filters(config_txt, peq_block_name, allow_empty=True)
    # before_value, after_value, filt_section = extract_value_section(config_txt, self.__block)
    # separate the tokens, which are in (TOKEN) blocks, from within the Value element
    if filt_element.text:
        filt_fragments = [v + ')' for v in filt_element.text.split(')') if v]
        if len(filt_fragments) < 2:
            raise ValueError('Invalid input file - Unexpected <Value> format')
        individual_filters = [d for d in [item_to_dicts(f) for f in filt_fragments[2:]] if d]
        start_beq_idx = -1
        end_beq_idx = -1
        beq_name = ''
        for i, t in enumerate(individual_filters):
            if start_beq_idx > -1 and end_beq_idx > -1:
                break
            f = t[0]
            if f.get('Type', '') == '20':
                txt = f.get('Text', '')
                if txt.startswith('***BEQ_START|'):
                    start_beq_idx = i
                    beq_name = txt[13:]
                if txt.startswith('***BEQ_END|'):
                    end_beq_idx = i
                    if txt[11:] != beq_name:
                        raise ValueError(f'Unexpected BEQ names {beq_name} vs {txt[11:]}')
        if start_beq_idx > -1:
            if end_beq_idx > start_beq_idx:
                new_filts = [f[1] for idx, f in enumerate(individual_filters) if idx < start_beq_idx or idx > end_beq_idx]
                new_filt_count = len(new_filts)
                filt_fragments[1] = f"({len(str(new_filt_count))}:{new_filt_count})"
                filt_element.text = ''.join(filt_fragments[0:2]) + ''.join(new_filts)
                config_txt = et.tostring(root, encoding='UTF-8', xml_declaration=True).decode('utf-8')
                return config_txt
            else:
                raise ValueError(f'Unexpected BEQ format {end_beq_idx} <= {start_beq_idx}')
        else:
            logger.info("No BEQ filter found")
            return None

    else:
        logger.warning("No filter txt found in DSP config")
        return None


def get_peq_key_name(block):
    '''
    :param block: 0 or 1.
    :return: the PEQ key name.
    '''
    if block == 0:
        return 'Parametric Equalizer'
    elif block == 1:
        return 'Parametric Equalizer 2'
    else:
        raise ValueError(f"Unknown PEQ block {block}")


class MediaServer:

    def __init__(self, ip: str, auth: Optional[Tuple[str, str]] = None, secure: bool = False):
        self.__ip = ip
        self.__auth = auth
        self.__secure = secure
        self.__base_url = f"http{'s' if secure else ''}://{ip}/MCWS/v1"
        self.__token = None

    def as_dict(self) -> dict:
        return {self.__ip: (self.__auth, self.__secure)}

    def __repr__(self):
        suffix = f" [{self.__auth[0]}]" if self.__auth else ' [Unauthenticated]'
        return f"{self.__ip}{suffix}"

    def authenticate(self) -> bool:
        self.__token = None
        url = f"{self.__base_url}/Authenticate"
        r = requests.get(url, auth=self.__auth, timeout=(1, 5))
        if r.status_code == 200:
            response = et.fromstring(r.content)
            if response:
                r_status = response.attrib.get('Status', None)
                if r_status == 'OK':
                    for item in response:
                        if item.attrib['Name'] == 'Token':
                            self.__token = item.text
        if self.connected:
            return True
        else:
            raise MCWSError('Authentication failure', r.url, r.status_code, r.text)

    @property
    def connected(self) -> bool:
        return self.__token is not None

    def get_zones(self) -> Dict[str, dict]:
        self.__auth_if_required()
        r = requests.get(f"{self.__base_url}/Playback/Zones", params={'Token': self.__token}, timeout=(1, 5))
        if r.status_code == 200:
            response = et.fromstring(r.content)
            if response:
                r_status = response.attrib.get('Status', None)
                if r_status == 'OK':
                    zones: Dict[str, dict] = {}
                    remote_zones = []
                    for child in response:
                        if child.tag == 'Item' and 'Name' in child.attrib:
                            attrib = child.attrib['Name']
                            if attrib.startswith('ZoneName'):
                                item_idx = attrib[8:]
                                if item_idx in zones:
                                    zones[item_idx]['name'] = child.text
                                else:
                                    zones[item_idx] = {'name': child.text}
                            elif attrib.startswith('ZoneID'):
                                item_idx = attrib[6:]
                                if item_idx in zones:
                                    zones[item_idx]['id'] = child.text
                                else:
                                    zones[item_idx] = {'id': child.text}
                            elif attrib.startswith('ZoneDLNA'):
                                if child.text == '1':
                                    remote_zones.append(attrib[8:])
                    return {v['id']: v for k, v in zones.items() if k not in remote_zones}
        raise MCWSError('No zones loaded', r.url, r.status_code, r.text)

    def get_zone_id(self, zone_name: str) -> str:
        zones = self.get_zones()
        return next(k for k, v in zones.items() if v['name'] == zone_name)

    def __auth_if_required(self):
        if not self.connected:
            self.authenticate()

    def get_dsp(self, zone_id: str) -> Optional[str]:
        self.__auth_if_required()
        r = requests.get(f"{self.__base_url}/Playback/SaveDSPPreset",
                         params={'Token': self.__token, 'Zone': zone_id, 'ZoneType': 'ID'},
                         timeout=(1, 5))
        if r.status_code == 200:
            response = et.fromstring(r.text)
            if response:
                if response.tag == 'DSP':
                    return r.text
                elif response.tag == 'Response':
                    r_status = response.attrib.get('Status', None)
                    if r_status == 'OK':
                        for child in response:
                            if child.tag == 'Item' and 'Name' in child.attrib and child.attrib['Name'] == 'Preset':
                                return child.text
        raise MCWSError('No DSP loaded', r.url, r.status_code, r.text)

    def set_dsp(self, zone_id: str, dsp: str) -> bool:
        self.__auth_if_required()
        dsp = dsp.replace('\n', '\r\n')
        if not dsp.endswith('\r\n'):
            dsp = dsp + '\r\n'
        r = requests.post(f"{self.__base_url}/Playback/LoadDSPPreset",
                          params={'Token': self.__token, 'Zone': zone_id, 'ZoneType': 'ID'},
                          files={'Name': (None, dsp)},
                          timeout=(1, 5))
        if r.status_code == 200:
            logger.info(f"LoadDSPPreset/{zone_id} success {r.url}")
            loaded_dsp = self.get_dsp(zone_id)
            if self.__compare_xml(et.fromstring(dsp), et.fromstring(loaded_dsp)):
                return True
            else:
                raise DSPMismatchError(zone_id, dsp, loaded_dsp)
        else:
            raise MCWSError('DSP not set', r.url, r.status_code, r.text)

    @staticmethod
    def __compare_xml(x1, x2):
        if x1.tag != x2.tag:
            return False
        for name, value in x1.attrib.items():
            if x2.attrib.get(name) != value:
                return False
        for name in x2.attrib:
            if name not in x1.attrib:
                return False
        if not MediaServer.__text_compare(x1.text, x2.text):
            return False
        if not MediaServer.__text_compare(x1.tail, x2.tail):
            return False
        cl1 = list(x1)
        cl2 = list(x2)
        if len(cl1) != len(cl2):
            return False
        i = 0
        for c1, c2 in zip(cl1, cl2):
            i += 1
            if not MediaServer.__compare_xml(c1, c2):
                return False
        return True

    @staticmethod
    def __text_compare(t1, t2):
        if not t1 and not t2:
            return True
        if t1 == '*' or t2 == '*':
            return True
        return (t1 or '').strip() == (t2 or '').strip()


class DSPMismatchError(Exception):

    def __init__(self, zone_id: str, expected: str, actual):
        super().__init__(f"Mismatch in DSP loaded to {zone_id}")
        self.zone_id = zone_id
        self.expected = expected
        self.actual = actual


class MCWSError(Exception):

    def __init__(self, msg: str, url: str, status_code: int, resp: Optional[str] = None):
        super().__init__(msg)
        self.msg = msg
        self.url = url
        self.status_code = status_code
        self.resp = resp


def item_to_dicts(frag) -> Optional[Tuple[Dict[str, str], str]]:
    idx = frag.find(':')
    if idx > -1:
        peq_xml = frag[idx+1:-1]
        vals = {i.attrib['Name']: i.text for i in et.fromstring(peq_xml).findall('./Item')}
        if 'Enabled' in vals:
            if vals['Enabled'] != '0' and vals['Enabled'] != '1':
                vals['Enabled'] = '1'
        else:
            vals['Enabled'] = '0'
        return vals, frag
    return None
